/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.components.jdbcquery;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.dbutils.DbUtils;
import org.apache.commons.dbutils.QueryRunner;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.bean.support.TransformBeanSupport;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.apache.servicemix.jbi.util.MessageUtil;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.dom.DOMDocument;
import org.dom4j.tree.DefaultDocument;
import org.springframework.jdbc.datasource.DataSourceUtils;

/**
 * Jdbc component that use the input NormalizedMessage to fill the query statement.
 * <br>
 * The result is optionally inserted into a copy of the incoming message otherwise
 * a new message is create as output.
 * 
 * @author buso
 *
 */
public class JDBCAdvancedQueryComponent extends TransformBeanSupport {
	
	private static final Log log = LogFactory.getLog(JDBCAdvancedQueryComponent.class);
	
	/**
	 * Information used to query the database
	 */
	private JDBCQueryConfig queryConfig;
	
	/**
	 * JDBC connection parameters
	 */
	private JDBCConnectionConfig connConfig;

	
	@Override
	protected boolean transform(MessageExchange exchange, NormalizedMessage in,
			NormalizedMessage out) throws Exception {

				
		// transform query in jdbc format
		List<String> orderedParameterNameList = getStringParameterNames(queryConfig.getQuery(), queryConfig.getMatchParamRegexp());
		
		String jdbcQuery = queryConfig.getQuery().replaceAll(queryConfig.getMatchParamRegexp(),"?");
		
		List<QueryParameterConfig> queryParametersConfig = queryConfig.getQueryParams();
		
		Map<String, QueryParameterConfig> parameterConfigMap = buildParameterConfigMap(queryParametersConfig);
		
		Object[] params = queryConfig.getParamsFromInXml(in, orderedParameterNameList,parameterConfigMap);
		
		// execute query
			// get connection
		Connection conn = connConfig.getConnection();
		Element results = null;
		try {
			QueryRunner runner = new QueryRunner();
			log.debug("Executing query: "+jdbcQuery);
			try {
				if (jdbcQuery.toLowerCase().indexOf("select") != -1) {
					results = (Element)runner.query(conn, jdbcQuery,
						params, new XmlResultSetHandler(
							queryConfig.getRowsXmlEnvelope(),
							queryConfig.getRowXmlEnvelope())
					);
				} else {
					int intResult = runner.update(conn, jdbcQuery,
							params);
					Element updateResult = DocumentHelper.createElement("affected-rows");
					updateResult.setText(""+intResult);
					results = updateResult;
				}
			} catch (SQLException sqle) {
				if (JDBCQueryConfig.FAULT_FLOW.equals(queryConfig.getFaultManagement())) {
					Fault fault = exchange.createFault();
						Element faultElem = DocumentHelper.createElement("Fault");
						Element sqlStateElem = DocumentHelper.createElement("sql-state");
						sqlStateElem.setText(sqle.getSQLState());
						Element errorCodeElem = DocumentHelper.createElement("error-code");
						errorCodeElem.setText(""+sqle.getErrorCode());
					faultElem.add(sqlStateElem);
					faultElem.add(errorCodeElem);
					fault.setContent(new StringSource(faultElem.asXML()));
					MessageUtil.transfer(fault, out);
					return true;
				}
				throw sqle;
			}
			log.debug("Query results: "+results);
		} finally {
			if (connConfig.getDatasource() != null) {
				DataSourceUtils.releaseConnection(conn, connConfig.getDatasource());
			} else {
				DbUtils.close(conn);
			}
		}
		
		
		// compose out response
		// 1) the results is a single row. The elements putted in some other dom element
		// 2) more rows. Must be setted as is
		
		MessageUtil.transfer(in, out);
		SourceTransformer trans = new SourceTransformer();
		String strContent = trans.contentToString(out);
		Document docContent = DocumentHelper.parseText(strContent);
		// search the envelop where to put the result
		Node contentEnvelop = null;
		
		if (isValidStr(queryConfig.getXmlEnvelope())){
			contentEnvelop = docContent.selectSingleNode(
					"/"+queryConfig.getXmlEnvelope());
			if (contentEnvelop == null) {
				log.warn("Envelope to enrich not found ["+"/"+queryConfig.getXmlEnvelope()+"]: " +
				"the result will be the original message");
			}
		}else{
			log.info("Xml Envelope is null we will keep the original message ");
		}
		
		
		
		

	
		if (contentEnvelop != null && queryConfig.isEnrichMessage() && isValidStr(queryConfig.getXmlEnvelope())) {
			//
			// Case 1: enrich message = true, xmlenvelop != null
			// Result is Original message + result attached as child of (queryConfig.getXmlEnvelope())
			//
			log.debug("Enriching Original Message, JDBC Result are children of queryConfig.getXmlEnvelope()");
			if (results.elements().size() == 1) {
				log.debug("Result has exactly 1 row");
				// the xpath on Element (results) didn't work (I do not know why)
				StringBuffer rowXpath = new StringBuffer("/")
					.append(queryConfig.getRowsXmlEnvelope()).append("/")
					.append(queryConfig.getRowXmlEnvelope());
				Node row = new DefaultDocument(results).selectSingleNode(rowXpath.toString());
				
				if (row != null) {
					log.debug("Adding result to envelope: "+row.getPath());
					((Element)contentEnvelop).appendContent( (Element)row );
				} else {
					log.warn("Envelope with single row result not found ["+rowXpath+"]: " +
						"all the results will be putted in: "+contentEnvelop.getPath());

					((Element)contentEnvelop).add(results);
				}
			} else {
				log.debug("Result has more rows; adding entire result to: "+contentEnvelop.getPath());
				((Element)contentEnvelop).add(results);
			}
			out.setContent(new StringSource( docContent.asXML() ));
		
		} else if (contentEnvelop == null && queryConfig.isEnrichMessage()){
			//
			// Case 2: enrich message = true, xmlenvelop == null
			// Result will be the original message
			//
			log.debug("Enriching =  true, xmlEnvelope is null -> Keep The Original Message");
			out.setContent(new StringSource(docContent.asXML()));
		} else {
			//
			// That is the case in which we do not keep original message but we create a new 
			//
			log.debug("Not enriching... Keep only the output result");
			docContent = new DOMDocument(); 
			docContent.add(results);
			out.setContent(new StringSource(docContent.asXML()));
		}
		
		return true;
	}

	public JDBCQueryConfig getQueryConfig() {
		return queryConfig;
	}

	public void setQueryConfig(JDBCQueryConfig config) {
		this.queryConfig = config;
	}

	public JDBCConnectionConfig getConnConfig() {
		return connConfig;
	}


	public void setConnConfig(JDBCConnectionConfig connConfig) {
		this.connConfig = connConfig;
	}
	
	public List<String> getStringParameterNames(String propertyValue, String matchRegexp) {
        Pattern pattern = Pattern.compile(matchRegexp);
        Matcher matcher = 
        	pattern.matcher(propertyValue);
        String placeHolder;
        List<String> toReturn = new ArrayList<String>();
		for (int idx = 1; matcher.find(); idx++) {
			placeHolder = matcher.group();
			placeHolder = placeHolder.substring(1);
			toReturn.add(placeHolder);
		}
		return toReturn;
	}
	
	private Map<String, QueryParameterConfig> buildParameterConfigMap(List<QueryParameterConfig> queryParametersConfig){
		Map<String, QueryParameterConfig> mapParamConfig = new HashMap<String, QueryParameterConfig>();
		for (QueryParameterConfig q : queryParametersConfig){
			mapParamConfig.put(q.getPlaceHolder(), q);
		}
		return mapParamConfig;
	}
	
	private boolean isValidStr(String s){
		return s != null && s.trim().length() > 0;
	}
	

}
