/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.components.restart;

import java.lang.reflect.Field;
import java.util.Map;

import javax.annotation.PostConstruct;
import javax.jbi.JBIException;
import javax.jbi.management.MBeanNames;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.MessageExchangeListener;
import org.apache.servicemix.bean.support.TransformBeanSupport;
import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.XPath;
import org.dom4j.xpath.DefaultXPath;

/**
 * Service Engine to resend a Message Exchange
 * It receives the id of the JournalInstanceService on the database, containing
 * the Message Exchange to resend.
 */
public class RestartComponent extends TransformBeanSupport implements
		MessageExchangeListener {

	// Single Restart expressions
	private static final String SINGLE_INSTANCE_RESTART_XPATH_EXPRESSION = "//restart:restartSingleInstance";
	
	// Massive Instance Restart expressions
	private static final String MASSIVE_INSTANCE_RESTART_XPATH_EXPRESSION = "//restart:restartMultipleInstances";

	public static final String CORRELATION_ID_PREFIX = "RESTARTED_ID:";
	
	private static final Log log = LogFactory.getLog(RestartComponent.class);

	private JBIContainer container = null;
	private Map<String, String> namespaceMap = null;
	
	// Generator for correlation ids
	private static IdGenerator idGenerator = null;

	public Map<String, String> getNamespaceMap() {
		return namespaceMap;
	}

	public void setNamespaceMap(Map<String, String> namespaceMap) {
		this.namespaceMap = namespaceMap;
	}
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}

	public void setContainer(JBIContainer container) {
		this.container = container;
	}
	
	@PostConstruct
	public void init() throws JBIException {

		try {
			// Retrieving JBI container
			// The method getContext() doesn't return a real Component Context but instead a
			// org.apache.servicemix.bean.BeanEndpoint$PojoContext, while we need
			// a org.apache.servicemix.jbi.framework.ComponentContextImpl
			// that is the object that contains the JBI container.
			// For this reason we retrieve the MBeanNames object, that in SMX 3.3
			// is the class ComponentContextImpl
			MBeanNames componentContextImpl = getContext().getMBeanNames();
			Field field = componentContextImpl.getClass().getDeclaredField("container");
			field.setAccessible(true);
			JBIContainer container = (JBIContainer) field.get(componentContextImpl);
			setContainer(container);
			log.info("Retrieved JBI container");
						
			// Create unique id generator
			if (idGenerator == null) {
				idGenerator = new IdGenerator(CORRELATION_ID_PREFIX);
			}
		} catch (Exception ex) {
			log.error("Unable to retrieve JBI container", ex);
			throw new JBIException(ex);
		}
	}
		
	/**
	 * Apply an XPath expression to the DOM document.
	 * @param document The DOM document of the message
	 * @return True if the expression select some node on the DOM
	 */
	private boolean checkExpression(Document document, String xPathExpression) {
		// According to the service type to activate (single/massive)
		// the input XML is different
		XPath genericXPathExpression = new DefaultXPath(xPathExpression);
		genericXPathExpression.setNamespaceURIs(getNamespaceMap());
		return (genericXPathExpression.selectSingleNode(document) != null);
	}
	
	/**
	 * Parse the message inside the exchange to get a DOM document.
	 * @param exchange The exchange containing the message
	 * @return The DOM document of the message
	 * @throws Exception
	 */
	private Document retrieveDOMDocument(NormalizedMessage in) throws Exception {
		// Create DOM for content
		SourceTransformer sourceTransformer = new SourceTransformer();
		String xmlMessage = sourceTransformer.contentToString(in);
		Document document = DocumentHelper.parseText(xmlMessage);

		return document;
	}
	
	@Override
    public NormalizedMessage getInMessage(MessageExchange exchange) throws org.apache.servicemix.jbi.exception.NoInMessageAvailableException  {
    	return super.getInMessage(exchange);
    }


	@Override
	protected boolean transform(MessageExchange exchange, NormalizedMessage in, NormalizedMessage out) throws Exception {
		debug("Received a request to restart one or more exchange");

		if (container == null) {
			log.error("Container not available: impossible restarting message");
			throw new MessagingException("Container not available: impossible restarting message");
		}
		
		try {
			// Parse the message to get a DOM, only once
			Document document = retrieveDOMDocument(in);
			
			// Check if we have to restart a single exchange or a list of exchanges
			if (checkExpression(document, SINGLE_INSTANCE_RESTART_XPATH_EXPRESSION)) {
				SingleRestart singleRestartManager = new SingleRestart(container, getNamespaceMap(), idGenerator, this);
				singleRestartManager.manageSingleRestart(exchange, document, out);
			} else if (checkExpression(document, MASSIVE_INSTANCE_RESTART_XPATH_EXPRESSION)) {
				MultipleInstanceRestart multipleInstanceRestartManager = new MultipleInstanceRestart(container, getNamespaceMap(), idGenerator, this);
				multipleInstanceRestartManager.manageMassiveRestart(exchange, document, out);
			} else {
				MultipleMessageRestart multipleMessageRestartManager = new MultipleMessageRestart(container, getNamespaceMap(), idGenerator, this);
				multipleMessageRestartManager.manageMassiveMessagesRestart(exchange, document, out);
			}

		} catch (Exception e) {
	    	log.error("Fault on restarting activity", e);
	    	// Throw the exception: the super class should send a Fault message or an error,
	    	// according to the MEP
	    	throw e;
		}
		
		return true;
	}

}
