package org.spagic.smx.components.restart;

import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.TransitionInstance;

import java.util.List;
import java.util.Map;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.servicemix.id.IdGenerator;
import org.apache.servicemix.jbi.container.JBIContainer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.dom4j.Document;
import org.dom4j.XPath;
import org.dom4j.xpath.DefaultXPath;
import org.hibernate.Session;

/**
 * Manages the restart of a single InstanceSA with the possibility of changing the messages.
 * The InstanceSA has to be faulted.
 *
 */
public class SingleRestart extends RestartBase {
	
	// Single Restart expressions
	private static final String ID_INSTANCE_XPATH_EXPRESSION = "//restart:restartSingleInstance/restart:idInstanceSA";
	private static final String SINGLE_RESTART_MODIFIED_MESSAGES_XPATH_EXPRESSION = "//restart:restartSingleInstance/restart:modifiedMessages/restart:modifiedMessage";
	
	public SingleRestart(JBIContainer container, Map<String, String> namespaceURI, IdGenerator idGenerator, RestartComponent restartComponent) {
		super(container, namespaceURI, idGenerator, restartComponent);
	}

	/**
	 * Extract from the DOM the id of the TransitionInstance containing
	 * the Exchange to resend
	 * @param document DOM to parse
	 * @return Id of the TransitionInstance
	 * @throws Exception
	 */
	private String retrieveIdInstanceSAToRestart(Document document) {
		// Retrieve Id of the instance to restart
		XPath idInstanceXPathExpression = new DefaultXPath(ID_INSTANCE_XPATH_EXPRESSION);
		idInstanceXPathExpression.setNamespaceURIs(getNamespaceURI());
		String extractedID = idInstanceXPathExpression.valueOf(document);
		return extractedID;
	}

	/**
	 * Extract from the DOM the id of the InstanceSA to restart
	 * @param document DOM to parse
	 * @return Id of the InstanceSA
	 */
	private long retrieveInstanceSAToRestart(Document document) {
		// Retrieve Id of the instance to restart
		return Long.parseLong(retrieveIdInstanceSAToRestart(document));
	}

	/**
	 * Perform a single setup.
	 * @param exchange Request exchange
	 * @param document The DOM document of the message
	 * @throws Exception
	 */
	public void manageSingleRestart(MessageExchange exchange, Document document, NormalizedMessage out) throws Exception {
		
		// Sample request message
		//		<restartSingleInstance xmlns="urn:it:eng:spagic:restart">
		//		   <idInstanceSA>123</idInstanceSA>
		//		   <modifiedMessages>
		//		      <modifiedMessage>
		//		         <idJournal>123</idJournal>
		//		         <xmlMessage>
		//		         .....
		//		         </xmlMessage>
		//		      </modifiedMessage>
		//		      <modifiedMessage>
		//		         <idJournal>456</idJournal>
		//		         <xmlMessage>
		//		         .......
		//		         </xmlMessage>
		//		      </modifiedMessage>
		//		   </modifiedMessages>
		//		</restartSingleInstance>
		
		
		// Sample response message without errors
		//		<restartSingleInstanceResponse xmlns="urn:it:eng:spagic:restart" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
		//		   <restartSingleInstanceReturn>
		//		      <originalInstanceId>54</originalInstanceId>
		//		      <restartedInstanceId>69</restartedInstanceId>
		//		      <errorMessage xsi:nil="true" />
		//		   </restartSingleInstanceReturn>
		//		</restartSingleInstanceResponse>
		//
		// Sample response message with errors
		//		<restartSingleInstanceResponse xmlns="urn:it:eng:spagic:restart">
		//		   <restartSingleInstanceReturn>
		//		      <originalInstanceId>54</originalInstanceId>
		//		      <restartedInstanceId>69</restartedInstanceId>
		//		      <errorMessage>simple error</errorMessage>
		//		   </restartSingleInstanceReturn>
		//		</restartSingleInstanceResponse>

		
		
		// Retrieve ID of message to restart
		long instanceSAId = retrieveInstanceSAToRestart(document);
		debug("Instance Id of Service Assembly retrieved: " + instanceSAId);
		
		// Retrieve all modified messages
		List<ModifiedMessage> modifiedMessages = retrieveModifiedMessages(document, SINGLE_RESTART_MODIFIED_MESSAGES_XPATH_EXPRESSION);
		debug("Number of modified messages: " + modifiedMessages.size());
		
		Session aSession = null;
		PrepareRestartResult prepareResult = null;
		
		try {
			try {
				aSession = HibernateUtil.getSessionFactory().openSession();
				
				prepareResult = prepareInstanceSARestart(aSession, instanceSAId);
				
				// Check for errors
				if (prepareResult.errorMessage != null) {
					throwErrorNotification(NO_INSTANCE_SA_CREATED, prepareResult.errorMessage);
				}
			}
			finally {
				if (aSession != null) {
					aSession.close();
				}
			}
					
			String newCorrelationId = prepareResult.processInstanceDuplicated.getIdCorrelation();
			for (TransitionInstance jis : prepareResult.transitionInstances) {
				String singleErrorMessage = restartSingleJIS(jis, newCorrelationId, modifiedMessages);
				if (singleErrorMessage != null) {
					throwErrorNotification(prepareResult.processInstanceDuplicated.getIdProcessInstance(), "Exchange found on database with wrong state. JIS id: " + jis.getIdJournal());
				}
			}
		}
		catch (RestartException ex) {
			// Send response, with errors
			out.setContent(new StringSource(prepareSingleRestartResponse(instanceSAId, ex.getRestartedInstanceId(), ex.getErrorMessage())));
			return;
		}
		
		// Send response, no errors
		out.setContent(new StringSource(prepareSingleRestartResponse(instanceSAId, prepareResult.processInstanceDuplicated.getIdProcessInstance(), null)));
    }
	
	
	private String prepareSingleRestartResponse(long instanceToRestartId, long restartedInstanceId, String errorMessage) {
		// Sample response message without errors
		//		<restartSingleInstanceResponse xmlns="urn:it:eng:spagic:restart" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
		//		   <restartSingleInstanceReturn>
		//		      <originalInstanceId>54</originalInstanceId>
		//		      <restartedInstanceId>69</restartedInstanceId>
		//		      <errorMessage xsi:nil="true" />
		//		   </restartSingleInstanceReturn>
		//		</restartSingleInstanceResponse>
		//
		// Sample response message with errors
		//		<restartSingleInstanceResponse xmlns="urn:it:eng:spagic:restart">
		//		   <restartSingleInstanceReturn>
		//		      <originalInstanceId>5</originalInstanceId>
		//		      <restartedInstanceId>69</restartedInstanceId>
		//		      <errorMessage>simple error</errorMessage>
		//		   </restartSingleInstanceReturn>
		//		</restartSingleInstanceResponse>
		//

		StringBuffer responseBuffer = new StringBuffer();
		responseBuffer.append("<restartSingleInstanceResponse xmlns=\"urn:it:eng:spagic:restart\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
		responseBuffer.append("<restartSingleInstanceReturn>");
		
		responseBuffer.append("<originalInstanceId>");
		responseBuffer.append(instanceToRestartId);
		responseBuffer.append("</originalInstanceId>");
		
		responseBuffer.append("<restartedInstanceId>");
		responseBuffer.append(restartedInstanceId);
		responseBuffer.append("</restartedInstanceId>");
		if (errorMessage != null) {
			responseBuffer.append("<errorMessage><![CDATA[");
			responseBuffer.append(errorMessage);
			responseBuffer.append("]]></errorMessage>");
		} else {
			responseBuffer.append("<errorMessage xsi:nil=\"true\" />");
		}
		responseBuffer.append("</restartSingleInstanceReturn>");
		responseBuffer.append("</restartSingleInstanceResponse>");
		return responseBuffer.toString();
	}
	
}
