/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.smx.components.sync;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;

import javax.annotation.PostConstruct;
import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.MessageExchangeListener;
import org.apache.servicemix.bean.support.BeanSupport;
import org.apache.servicemix.components.util.ComponentSupport;
import org.apache.servicemix.components.util.CopyTransformer;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.spagic.smx.components.restart.RestartComponent;
import org.spagic.smx.listeners.sync.SynchronizerFaultListener;


public class TransactedSynchronizer extends BeanSupport implements
		MessageExchangeListener {
	
	public static final String SYNCHRONIZER_SERVICE = "SYNCHRONIZER_SERVICE".intern();
	public static final String SYNCHRONIZER_ENDPOINT = "SYNCHRONIZER_ENDPOINT".intern();
	public static final String HIBERNATE_SESSION_PROPERTY = "org.spagic.transactions.currentSession".intern();
	
	private static final Log log = LogFactory.getLog(TransactedSynchronizer.class);

	private HashMap<String, MessageExchange> messageExchangeStore = null;
	private HashMap<String, Session> hibernateSessionStore = null;
    private boolean copyProperties = true;
    private boolean copyAttachments = true;
    private SessionFactory sessionFactory = null;
    
	public boolean isCopyAttachments() {
		return copyAttachments;
	}

	public void setCopyAttachments(boolean copyAttachments) {
		this.copyAttachments = copyAttachments;
	}

	public boolean isCopyProperties() {
		return copyProperties;
	}

	public void setCopyProperties(boolean copyProperties) {
		this.copyProperties = copyProperties;
	}

	/* (non-Javadoc)
	 * @see org.apache.servicemix.MessageExchangeListener#onMessageExchange(javax.jbi.messaging.MessageExchange)
	 */
	public void onMessageExchange(MessageExchange exchange)
			throws MessagingException {
		log.debug("Synchronizer::onMessageExchange");
		log.debug("Sender: "+exchange.getProperty(JbiConstants.SENDER_ENDPOINT));
		log.debug(exchange);

		// Retrieve correlation id
		String processCorrelationId = (String) exchange.getProperty(JbiConstants.CORRELATION_ID);
		if (processCorrelationId == null) {
			log.warn("Process Correlation ID not present in exchange: " + exchange);
			return;
		}
		
		if (isInAndOut(exchange)) {
			if (exchange.getStatus().equals(ExchangeStatus.DONE) ||
				exchange.getStatus().equals(ExchangeStatus.ERROR)) {
				// No further processing
				return;
			}
			manageInOutActive(exchange, processCorrelationId);
		} else {
			// InOnly or Robust InOnly exchange
			if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
				if (isRobustInOnly(exchange)) {
					// Robust InOnly: should send a DONE message
					done(exchange);
		            return;
				} else {
					// InOnly: no further processing
					return;
				}
			} else if (exchange.getError() == null) {
				manageInOnlyActive(exchange, processCorrelationId);
			} else {
				// Error
				manageInOnlyError(exchange, processCorrelationId);
			}
		}
	}
	
	private void manageInOutActive(MessageExchange exchange, String processCorrelationId) throws MessagingException {
		
		Transaction tx = null;
		Session aSession = null;

		// I must check if this is the request or the response
		if (exchange.getRole() == MessageExchange.Role.PROVIDER) {
			log.debug("InOut exchange, Role is PROVIDER ");
		} else if (exchange.getRole() == MessageExchange.Role.CONSUMER) {
			log.error("InOut exchange, Role is CONSUMER. This is a response of an IN-OUT exchange.");
			return;
		}

		
		// Opening a new Session on the configured session factory
		log.debug("Opening a new Session");
		aSession = sessionFactory.openSession();
		log.debug("Session retrieved: ");
		log.debug(aSession);
    	
		log.debug("Beginning a new transaction");
		tx = aSession.beginTransaction();
		
			
		QName serviceQName = getService();
		String endpoint = getEndpoint();

		// Retrieve input message for input-output MEP
		NormalizedMessage inMessageForInOutExchange = getInMessage(exchange);
		InOnly exchangeToSend = getExchangeFactory().createInOnlyExchange();

		// Configure the In-Only exchangeTarget
//		target.configureTarget(exchangeToSend, getContext());

		// Set the properties used by the monitoring components
		exchangeToSend.setProperty(JbiConstants.SENDER_ENDPOINT, serviceQName + ":" + endpoint);
		exchangeToSend.setProperty(JbiConstants.CORRELATION_ID, processCorrelationId);

		// Create message for In-Only MEP
		NormalizedMessage inMessageForInOnlyExchange = exchangeToSend.createMessage();
		exchangeToSend.setMessage(inMessageForInOnlyExchange, "in");
		
		// Copy content of input message to content of In-Only input message
		inMessageForInOnlyExchange.setContent(inMessageForInOutExchange.getContent());
		
		// Set some properties in the message allowing to recognize that the
		// Synchronizer is involved
		inMessageForInOnlyExchange.setProperty(SYNCHRONIZER_SERVICE, serviceQName);
		inMessageForInOnlyExchange.setProperty(SYNCHRONIZER_ENDPOINT, endpoint);
		

		// Store the Hibernate session in a message property, for later use
		inMessageForInOnlyExchange.setProperty(HIBERNATE_SESSION_PROPERTY, aSession);

		
		// If required, copy also attachments and properties
        copyPropertiesAndAttachments(inMessageForInOutExchange, inMessageForInOnlyExchange);

		// Store message for later use
		log.debug("Inserted message with correlation id :" + processCorrelationId + " in message store");
		messageExchangeStore.put(processCorrelationId, exchange);
		hibernateSessionStore.put(processCorrelationId, aSession);

		// Send In-Only input message
		getDeliveryChannel().send(exchangeToSend);
	}
	
	private void manageInOnlyActive(MessageExchange exchange, String processCorrelationId) throws MessagingException {
		// I must check if this is the request or the response
		if (exchange.getRole() == MessageExchange.Role.PROVIDER) {
			log.debug("InOnly exchange, Role is PROVIDER ");
		} else if (exchange.getRole() == MessageExchange.Role.CONSUMER) {
			log.error("InOnly exchange, Role is CONSUMER. This is a response of an IN-ONLY exchange.");
			return;
		}

		// The exchange is InOnly: in the store we should find the corresponding InOut exchange
		NormalizedMessage inMessageForInOnlyExchange = getInMessage(exchange);
		
		// Retrieve In-Out Message Exchange
		log.debug("Searching message with correlation id :" + processCorrelationId + " in message store");
		MessageExchange inOutMessageExchange = messageExchangeStore.get(processCorrelationId);
		
		if (inOutMessageExchange == null) {
			
			// Check if the exchange is for a restart: in that case it's normal not finding
			// the InOut exchange in our map, so we have only to send the DONE
			if (processCorrelationId.startsWith(RestartComponent.CORRELATION_ID_PREFIX)) {
				log.debug("Unable to find In-Out Message Exchange for In-Only exchange: " + exchange);
		        done(exchange);
			} else {
				String errorMessage = "Unable to find In-Out Message Exchange for In-Only exchange: " + exchange;
				log.error(errorMessage);
				
				Fault fault = exchange.createFault();
				fault.setContent(generateXMLError(errorMessage, null));
				fail(exchange, fault);
			}
			return;
			
		}
		
		// Retrieve content from InOnly exchange and copy it to the InOut exchange
		NormalizedMessage outMessageForInOutExchange = inOutMessageExchange.createMessage();
		inOutMessageExchange.setMessage(outMessageForInOutExchange, "out");
		outMessageForInOutExchange.setContent(inMessageForInOnlyExchange.getContent());
		
		// If required, copy also attachments and properties
        copyPropertiesAndAttachments(inMessageForInOnlyExchange, outMessageForInOutExchange);
		
		getDeliveryChannel().send(inOutMessageExchange);
		

		// According to the content of the message, we commit or rollback the transaction
		Session aSession = hibernateSessionStore.get(processCorrelationId);
		Transaction tx = aSession.getTransaction();
		try {
			SourceTransformer sourceTransformer = new SourceTransformer();
			String xmlInput = sourceTransformer.contentToString(inMessageForInOnlyExchange);
			if (xmlInput.indexOf("<errorMessage>") == -1) {
				log.debug("Committing the transaction");
				tx.commit();
			} else {
				log.debug("Rolling back the transaction");
				tx.rollback();
			}
		} catch (Exception ex) {
			if (tx.isActive()) {
				tx.rollback();
			}
			log.error(ex);
		} finally {
			aSession.close();
		}
		
		
		log.debug("Removed message with correlation id :" + processCorrelationId + " in message store");
		messageExchangeStore.remove(processCorrelationId);
		hibernateSessionStore.remove(processCorrelationId);

		// Send DONE message for InOnly exchange
		if (exchange.getProperty(SynchronizerFaultListener.SYNCHRONIZER_FAULT_LISTENER_PROPERTY) == null) {
			// The original exchange was not sent by a listener
	        done(exchange);
		}
	}
	
	/**
	 * Generates an XML error containing an optional exception stack trace
	 * @param errorMessage Message to trace
	 * @param error Optional exception to trace
	 * @return XML error message
	 */
	private StringSource generateXMLError(String errorMessage, Exception error) {
		
		StringWriter errorCause = new StringWriter();
		errorCause.write("<error><message>");
		errorCause.write(errorMessage);
		errorCause.write("</message>");
		if (error != null) {
			errorCause.write("<stackTrace><![CDATA[");
			error.printStackTrace(new PrintWriter(errorCause));
			errorCause.write("]]></stackTrace>");
		}
		errorCause.write("</error>");
		return new StringSource(errorCause.toString());
		
	}
	
	private void manageInOnlyError(MessageExchange exchange, String processCorrelationId) throws MessagingException {
		// For Robust InOnly should we produce a DONE message ?
		Source errorContent = null;
		if (exchange.getError() != null) {
			errorContent = generateXMLError(exchange.getError().getMessage(), exchange.getError());
        } else if (exchange.getFault() != null) {
        	errorContent = exchange.getFault().getContent();
        }
						
		// Retrieve In-Out Message Exchange
		log.debug("Searching message with correlation id :" + processCorrelationId + " in message store");
		MessageExchange inOutMessageExchange = messageExchangeStore.get(processCorrelationId);
		
		if (inOutMessageExchange == null) {
			log.error("Unable to find In-Out Message Exchange for In-Only exchange: " + exchange);
			return;
		}
		
		
		log.debug("Rolling back the transaction");
		Session aSession = hibernateSessionStore.get(processCorrelationId);
		Transaction tx = aSession.getTransaction();
		tx.rollback();
		
		
		
		log.debug("Removed message with correlation id :" + processCorrelationId + " in message store");
		messageExchangeStore.remove(processCorrelationId);
		hibernateSessionStore.remove(processCorrelationId);
		
		// Retrieve content from InOnly exchange and copy it to the InOut exchange
//		NormalizedMessage outMessageForInOutExchange = inOutMessageExchange.createMessage();
		NormalizedMessage outMessageForInOutExchange = inOutMessageExchange.createFault();
//		inOutMessageExchange.setMessage(outMessageForInOutExchange, "out");
		inOutMessageExchange.setMessage(outMessageForInOutExchange, "fault");
		outMessageForInOutExchange.setContent(errorContent);
		getDeliveryChannel().send(inOutMessageExchange);
	}

	public String getMep(MessageExchange exchange) {
		String mepClassName = exchange.getClass().getName();
		return mepClassName;
	}

	@PostConstruct
	public void init() throws JBIException {
		messageExchangeStore = new HashMap<String, MessageExchange>();
		hibernateSessionStore = new HashMap<String, Session>();
	}

    /**
     * A helper method which will return true if the exchange is capable of Robust InOnly.
     *
     * @param exchange
     * @return true if the exchange can handle robust InOnly
     */
    protected boolean isRobustInOnly(MessageExchange exchange) {
        return exchange instanceof RobustInOnly;
    }

    /**
     * If enabled the properties and attachments are copied to the destination message
     */
    protected void copyPropertiesAndAttachments(NormalizedMessage in, NormalizedMessage out) throws MessagingException {
        if (isCopyProperties()) {
            CopyTransformer.copyProperties(in, out);
        }
        if (isCopyAttachments()) {
            CopyTransformer.copyAttachments(in, out);
        }
    }

	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}

	public void setSessionFactory(SessionFactory sessionFactory) {
		this.sessionFactory = sessionFactory;
	}
    
}
